# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-29 15:48
Desc:
'''
import pandas as pd
import dill as pickle
from pynovice.score_card.src.data_boxing import frequence_blocks,distince_blocks, kmeans_blocks, bayesian_blocks, \
    ks_blocks, chi_blocks, tree_blocks

from pynovice.score_card.feature_preprocessing import FeatureGenerator

def test_data_binning():
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 6, 8, 3, 2, 1, 5, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field', 'label']
    df_field, df_label = df['field'], df['label']

    # frequence_blocks
    fre_blocks = frequence_blocks(x=df_field, box_num=4)
    print("frequence_blocks: {}".format(fre_blocks))

    # distince_blocks
    dist_blocks = distince_blocks(x=df_field, box_num=4)
    print("distince_blocks: {}".format(dist_blocks))

    # kmeans_blocks
    km_blocks = kmeans_blocks(x=df_field, box_num=4)
    print("kmeans_blocks: {}".format(km_blocks))

    # bayesian_blocks
    bay_blocks = bayesian_blocks(df_field, fitness='events', p0=0.01)
    print("bayesian_blocks: {}".format(bay_blocks))

    # ks_blocks
    k_blocks = ks_blocks(df_field, df_label, box_num=4)
    print("ks_blocks: {}".format(k_blocks))

    # chi_blocks
    ch_blocks = chi_blocks(df_field, df_label, box_num=4)
    print("chi_blocks: {}".format(ch_blocks))

    # cart_blocks
    cart_blocks = tree_blocks(df_field, df_label, box_num=4)
    print("tree_blocks: {}".format(cart_blocks))
    aa=pd.cut(df_field,cart_blocks)
    print(aa)
    # bb,cc=calc_woe(aa, df_label, bad_target=1, eps=1e-4)
    # print(bb,cc)



    # ## 类别变量转换
    # print("-----类别-----")
    # cate_coder = CateCoder()
    # df_field = cate_coder.cate_coding(df_field, df_label, target_label=1)
    # def get_blocks(df_field,df_label):
    #     fre_blocks = frequence_blocks(x=df_field, bins=4)
    #     dist_blocks = distince_blocks(x=df_field, bins=4)
    #     km_blocks = kmeans_blocks(x=df_field, bins=4)
    #     bay_blocks = bayesian_blocks(df_field, fitness='events', p0=0.01)
    #     k_blocks = ks_blocks(df_field, df_label, box_num=4)
    #     ch_blocks = chi_blocks(df_field, df_label, box_num=4)
    #     cart_blocks = tree_blocks(df_field, df_label, box_num=4)
    #     return (fre_blocks,dist_blocks,km_blocks,bay_blocks,k_blocks,ch_blocks,cart_blocks)
    # set_blocks = get_blocks(df_field,df_label)
    # for i,block in enumerate(set_blocks):
    #     block_dict = cate_coder.blocks_2_catedict(block)
    #     print(i,block_dict)

def test_feature_procprocessing():
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 6, 8, 3, 2, 1, 5, 7], [10, 2, 3, 42, 534, 5, 53, 6, 83, 3, 42, 21, 25, 7],
                       [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field1', 'field2', 'label']
    df_fields, df_label = df[['field1', 'field2']], df['label']

    feature_generator = FeatureGenerator()
    feature_generator.fit_transform(df_fields, df_label)
    print(feature_generator.binning_woe_info)
    file_name = '../data/feature_generater.pkl'
    pickle.dump(feature_generator, open(file_name, "wb"),protocol=3)

    # load
    feature_generator= pickle.load(open(file_name, "rb"))
    fields_json = {'field1': 4, 'field22': 5, 'aa': 9}
    feature = feature_generator.transform(fields_json)
    print(feature)

if __name__ == '__main__':
    # 分箱
    test_data_binning()
    test_feature_procprocessing()


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-23 16:10
Desc:
    聚类分箱(k-means) 合并
    >>> kmeans_blocks(x=[1,2,3],bins=3)
    >>> print(aa)
'''

from sklearn.cluster import KMeans
from sklearn.metrics import calinski_harabaz_score
import numpy as np

def kmeans_blocks(x,box_num=3):
    '''
        聚类分箱
    :param x: dtype=[]
    :param bins: 分箱数量
    :return: 箱边界
    '''
    len_clocks = min(box_num,len(x),len(set(x))+1)
    if len_clocks<=1:
        return [-np.inf,np.inf]
    X = np.array(x).reshape([-1, 1])
    km = KMeans(n_clusters=len_clocks-1, random_state=666)
    y_pre = km.fit_predict(X)
    # 使用Calinski-Harabasz Index评估的聚类分数: 分数越高，表示聚类的效果越好
    if km.cluster_centers_.size > 1:  # 聚类类别大于1
        kmeans_score = calinski_harabaz_score(X, y_pre)
        print("聚类效果评分：{}".format(kmeans_score))
    tb = km.cluster_centers_.reshape([-1])
    tb.sort()
    blocks = np.concatenate([[-np.inf],tb,[np.inf]])
    return blocks.tolist()

if __name__ == '__main__':
    aa = kmeans_blocks(x=[1, 2, 3], box_num=3)
    print(aa)


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-03-27 15:17
Desc:
'''

import sys
import math
import inspect
import signal

# 1.进度条展示 progress_bar(1, 100)
def progress_bar(portion, total):
    """
    total 总数据大小，portion 已经传送的数据大小
    :param portion: 已经接收的数据量
    :param total: 总数据量
    :return: 接收数据完成，返回True
    """
    part = total / 50  # 1%数据的大小
    count = math.ceil(portion / part)
    sys.stdout.write('\r')
    sys.stdout.write(('[%-50s]%.2f%%' % (('>' * count), portion / total * 100)))
    sys.stdout.flush()

    if portion >= total:
        sys.stdout.write('\n')
        return True
    return False

# 2.获取变量名
def get_retrieve_name(var):
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    list_name = [var_name for var_name, var_val in callers_local_vars if var_val is var]
    return list_name

# 3.函数超时退出
def set_timeout(seconds, callback_func):
    '''
    暂时不可用
    :param seconds: 超时时间秒
    :param callback_func:超时后执行的函数
    :return:
    '''
    def wrap(func):
        def handle(signum, frame):  # 收到信号 SIGALRM 后的回调函数，第一个参数是信号的数字，第二个参数是the interrupted stack frame.
            raise RuntimeError

        def to_do(*args, **kwargs):
            try:
                signal.signal(signal.SIGALRM, handle)  # 设置信号和回调函数
                signal.alarm(seconds)  # 设置 num 秒的闹钟
                # print('start alarm signal.')
                r = func(*args, **kwargs)
                # print('close alarm signal.')
                signal.alarm(0)  # 关闭闹钟
                return r
            except RuntimeError as e:
                if callback_func:
                    callback_func()

        return to_do
    return wrap

if __name__ == '__main__':
    ## 1.进度条
    portion = 0
    total = 254820000
    while True:
        portion += 1024
        sum = progress_bar(portion, total)
        if sum:
           break
    print("ok")
    # or
    progress_bar(portion=32, total=100)


    ## 2.变量名
    name = "bob"
    age = "23"
    bb = (name,age)
    for i in bb:
        var = get_retrieve_name(i)
        print(var)

    ## 3.函数超时
    def aa():
        print("超时")
    @set_timeout(2, callback_func=aa)
    def tt1():
        import time
        time.sleep(3)
    tt1()



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-22 14:26
Desc:
    评分卡流程(计算得分及各种评估)

    TPR = TP / (TP+FN);
    FPR = FP / (FP + TN);
'''
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

class ScoreCardModel:
    def __init__(self):
        self.load_score_alpha()

    def load_score_alpha(self, odds=1 / 2, bScore=650, addScore=-20):
        '''
            1.坏好比odds=1:2时，对应基础评分bScore=600
            2.坏好比增加一倍时，评分减少addscore=-20
            bScore = offset - factor*ln(odds)
            bScore + addScore = offset - factor*ln(2*odds)
        :param odds: 坏好比
        :param bScore: 基础score分
        :param addScore: 评分变动值
        :return: bScore = offset - factor*ln(odds)
        '''

        self.score_factor = addScore / np.log(2)
        self.score_offset = bScore - addScore * np.log(odds) / np.log(2)

    def probability_to_score(self, bad_prob, eps=1e-4):
        '''
            逾期概率转换为评分
        :param bad_prob: 逾期概率
        :return: bScore = offset - factor*ln(odds)
        '''

        odds = max(bad_prob, eps) / max(1 - bad_prob, eps)
        score = self.score_offset + self.score_factor * np.log(odds)
        return int(score)

    def score_to_probability(self, score, eps=1e-4):
        '''
            评分转换为逾期概率
        :param score: 信用评分
        :return: bScore = offset - factor*ln(odds) -->bad_prob
        '''
        score_factor = eps if self.score_factor==0 else self.score_factor
        odds = np.e**((score - self.score_offset)/(score_factor))
        odds = -1+eps if odds == -1 else odds
        bad_prob = 1.00*odds/(1+odds)
        return bad_prob

    @classmethod
    def get_auc(cls, df_pre, df_label, pre_target=1):
        '''
        功能: 计算KS值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_label: 一维数组或series，代表真实的标签{0,1}
        :return: 'auc': auc值，'crossdens': TPR&FPR
        '''
        df_label = df_label.apply(lambda x: 1 if x == pre_target else 0)
        crossfreq = pd.crosstab(df_pre, df_label).sort_index(ascending=False)  # 按照预测概率从大到小排序作为阈值
        crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        fpr = crossdens.loc[:, 'fpr']
        tpr = crossdens.loc[:, 'tpr']
        auc = np.trapz(y=tpr, x=fpr)
        return auc, crossdens

    @classmethod
    def get_ks(cls, df_pre, df_label, pre_target=1):
        '''
        功能: 计算KS值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_label: 一维数组或series，代表真实的标签
        :return:
            'ks': KS值
            'crossdens': 好坏客户累积概率分布以及其差值gap
        '''
        df_label = df_label.apply(lambda x: 1 if x == pre_target else 0)
        crossfreq = pd.crosstab(df_pre, df_label).sort_index(ascending=False)  # 按概率降序排序
        crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
        crossdens.columns = ['fpr', 'tpr']
        crossdens.name = 'pre_threshold'
        crossdens['gap'] = abs(crossdens['fpr'] - crossdens['tpr'])
        ks = crossdens[crossdens['gap'] == crossdens['gap'].max()]
        return ks, crossdens

    @classmethod
    def get_lift(cls, df_pre, df_label, pre_target=1, judg_threshold=0.5):
        '''
        功能: 计算KS值，输出对应分割点和累计分布函数曲线图
        :param df_pre: 一维数组或series，代表模型得分
        :param df_label: 一维数组或series，代表真实的标签{0,1}
        :return: 'auc': auc值，'crossdens': TPR&FPR
        '''
        df_label = df_label.apply(lambda x: 1 if x == pre_target else 0)
        crossfreq = pd.crosstab(df_pre, df_label).sort_index(ascending=False)
        crossdens = crossfreq.cumsum(axis=0)
        crossdens = crossdens.apply(lambda x: x / crossdens.sum(axis=1), axis=0)
        crossdens.columns = ['acc_false', 'acc_true']
        crossdens.name = 'pre_threshold'
        crossdens['acc_random'] = df_label.sum() / df_label.size  # 随机情况下的准确率
        crossdens['lift'] = crossdens['acc_true'] / crossdens['acc_random']
        # 预测阈值
        lift = crossdens[crossdens.index >= judg_threshold]['lift'].values[-1]
        return lift, crossdens

    @classmethod
    def plot_roc(cls, df_pre, df_label, pre_target=1, save_path='./'):
        auc, crossdens = cls.get_auc(df_pre, df_label, pre_target)
        fpr = crossdens.loc[:, 'fpr']
        tpr = crossdens.loc[:, 'tpr']

        print('auc=%0.5f' % auc)
        fig = plt.figure(figsize=(10, 10))
        plt.plot(fpr, tpr, 'r--', linewidth=2.0, aa=False, label='ROC (area=%0.2f)' % (auc))
        plt.plot([0, 1], [0, 1], '--', color=(0.6, 0.6, 0.6))
        plt.xlim([0.0, 1.00])
        plt.ylim([0.0, 1.00])
        plt.xlabel('False Postive Rate')
        plt.ylabel('True Postive Rate')
        plt.title('ROC Curve')
        plt.legend(loc="lower right")
        # plt.savefig('%s/roc_curve_v1.png' % (save_path), dpi=180)
        fig.savefig('%s/roc_curve_v1.png' % (save_path,), dpi=180)
        plt.close(fig)
        return auc

    @classmethod
    def plot_ks(cls, df_pre, df_label, pre_target=1, save_path='./'):
        ks, crossdens = cls.get_ks(df_pre, df_label, pre_target)
        max_ks_gap_index = str(ks.index[0])
        max_ks_gap_good_value = ks['tpr'].values[0]
        max_ks_gap_bad_value = ks['fpr'].values[0]
        max_ks_gap_value = ks['gap'].values[0]
        annotate_xtext = max_ks_gap_index
        annotate_ytext = str(round(max_ks_gap_value, 3))
        annotate_text_y_index = max_ks_gap_value / 2 + min(max_ks_gap_good_value, max_ks_gap_bad_value)

        fpr = crossdens.loc[:, 'fpr']
        tpr = crossdens.loc[:, 'tpr']
        axes_x = [str(i) for i in fpr.index]

        print('ks=%s' % max_ks_gap_value)
        fig = plt.figure(figsize=(10, 10))
        axes = fig.gca()
        axes.plot(axes_x, tpr, 'g', linewidth=2, label='tpr')
        axes.plot(axes_x, fpr, 'r', linewidth=2, label='fpr')
        axes.annotate(annotate_xtext, xy=(max_ks_gap_index, 0), xytext=(max_ks_gap_index, 0.05),
                      arrowprops=dict(facecolor='red', shrink=0.05))
        axes.plot([max_ks_gap_index, max_ks_gap_index],
                  [max_ks_gap_bad_value, max_ks_gap_good_value],
                  linestyle='--', linewidth=2.5)
        axes.annotate(annotate_ytext, xy=(max_ks_gap_index, annotate_text_y_index))
        axes.legend()
        axes.set_title('KS Curve')
        fig.savefig('%s/ks_curve_v1.png' % (save_path), dpi=180)
        plt.close(fig)
        return max_ks_gap_value

    @classmethod
    def plot_lift(cls, df_pre, df_label, pre_target=1, save_path='./'):
        lift, crossdens = cls.get_lift(df_pre, df_label, pre_target)
        crossdens['acc_random'] = 1
        acc_random = crossdens.loc[:, 'acc_random']
        acc_lift = crossdens.loc[:, 'lift']
        axis_x = [str(i) for i in acc_lift.index]

        print('lift=%0.5f' % lift)
        fig = plt.figure(figsize=(10, 10))
        plt.plot(axis_x, acc_random, 'g', linewidth=2.0, aa=False, label='random')
        plt.plot(axis_x, acc_lift, 'r', linewidth=2.0, aa=False, label='model(lift_threshold): %0.2f' % lift)

        plt.xlabel('Pre Rate')
        plt.title('Lift Curve')
        plt.legend(loc="lower right")
        fig.savefig('%s/lift_curve_v1.png' % (save_path,), dpi=180)
        plt.close(fig)
        return lift


if __name__ == '__main__':
    score_card_model = ScoreCardModel()
    score_card_model.load_score_alpha()
    for i in range(0, 11):
        score = score_card_model.probability_to_score(0.1 * i)
        print('overdue prob: {}, score: {}'.format(round(0.1 * i, 2), score))

    print('2343',score_card_model.score_to_probability(660))
    print(score_card_model.probability_to_score(0.3))

    # 评分卡模型评估
    df_raw = pd.DataFrame([[0.42, 0.73, 0.55, 0.37, 0.57, 0.70, 0.25, 0.23, 0.46, 0.62, 0.76, 0.46, 0.55, 0.56, 0.56,
                            0.38, 0.37, 0.73, 0.77, 0.21],
                           [1, 1, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0]]).T
    df_pre = df_raw[0]
    df_label = df_raw[1]

    # score_card_model.plot_roc(df_pre, df_label, pre_target=1, save_path='../data')
    # score_card_model.plot_ks(df_pre, df_label, pre_target=1, save_path='../data')
    # score_card_model.plot_lift(df_pre, df_label, pre_target=1, save_path='../data')


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-24 10:54
Desc:
    简易的风控规则引擎

    待系统化：（单量预警系统）
    |当前进件单量| 当前进件通过率| 当前通过单量| - |预测剩余进件量| 预测剩余进件通过率(7天同时段同规则)| 预测剩余通过单量|- |预测今日总计通过单量
    单量预测可靠性 = 当天已进件通过率 - 7天进件通过率(同时段同规则) 可靠性评分
'''

from .rule_parser import RuleParser


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-16 18:09
Desc:
map extends
'''
import webbrowser
import os
import numpy as np
import reverse_geocoder as rg
import pandas as pd
from pyecharts.datasets import register_url
from pyecharts import options as opts
from pyecharts.charts import Geo,Map,Tab
from pyecharts.globals import ChartType
import json
from pyecharts.components import Table
CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))
class GeoMap:
    def __init__(self, country='肯尼亚'):
        self.commit = "reverse Geocoder takes a latitude / longitude coordinate and returns the nearest town/city."
        self.df_geo_dict = None
        self.country = country
        self.series_name = country
        self.html_name = os.path.abspath('geo.html')
        self.echarts_url = "https://echarts-maps.github.io/echarts-countries-js/"
        self.geo_dict = {}
        self.init_assert()

    def init_assert(self):
        city_list = self.get_city_list()['CITY(key)'].to_list()
        if '中国' == self.country:
            self.country = 'china'
        elif '印度尼西亚' == self.country:
            self.country = '印度尼西亚, 印尼'
        elif self.country not in city_list:
            print('Get more country|city info, use GeoMap.get_city_list()... ')
        else:
            pass

    def set_out_file(self,html_path):
        self.html_name = os.path.abspath(html_path)

    def get_geo_reverse(self, df_raw):
        '''
        根据经纬度获取 城市信息
        :param df_raw: Dataframe containing latitude and longitude information
        :return: the nearest town/city info
        '''
        assert 'latitude' in df_raw.columns, 'assert latitude in columns'
        assert 'longitude' in df_raw.columns, 'assert longitude in columns'
        df_gps = df_raw[['latitude', 'longitude']].reset_index(drop=True,inplace=False)
        _tuple_gps = df_gps.apply(lambda x: (x['latitude'], x['longitude']), axis=1).to_list()
        _geo_raw = rg.search(_tuple_gps)
        _geo = pd.read_json(json.dumps(_geo_raw))
        df_geo = pd.concat([df_gps, _geo], axis=1)
        self.geo_dict = dict(df_geo.apply(lambda x:(x['admin1'],[x['lon'],x['lat']]),axis=1).values)
        return df_geo

    def plot_geo_summary(self,_geo_data,series_name=None,threshold_plot=None ):
        '''
        plot geo 经纬度聚类信息
        :param geo_data: geo对应的城市分布(Series)
        :return:
        '''
        self.series_name = series_name if series_name else self.series_name

        geo_data = np.array(list(dict(_geo_data).items()))
        self.threshold_plot = np.percentile(_geo_data, 80) if not threshold_plot else threshold_plot
        _geo = self.get_geo_base(geo_data)
        _map = self.get_map_base(geo_data)
        table = self.get_table_base(geo_data)

        tab = Tab(page_title=self.country, )  # 选项卡多图
        tab.add(_map, "VisualMap")
        tab.add(_geo, "HeatMap")
        tab.add(table, "Data")
        tab.render(self.html_name)

    def get_geo_base(self, _data):
        '''
        plot geo-HeatMap
        :param _data: gps对应的城市分布(dataframe)
        :return: geo城市分布图
        '''
        register_url(self.echarts_url)
        geo = Geo(init_opts=opts.InitOpts(width="1000px", height="600px", page_title=self.country, bg_color=''))
        geo.add_schema(maptype=self.country)
        for _city,_gps in self.geo_dict.items():
            geo.add_coordinate(name=_city, longitude=_gps[0], latitude=_gps[1])
        geo.add(series_name = self.series_name, data_pair = _data, type_=ChartType.HEATMAP)
        geo.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
        geo.set_global_opts(visualmap_opts=opts.VisualMapOpts(max_= self.threshold_plot),
                            title_opts=opts.TitleOpts(title="Geo-HeatMap"))
        return geo

    def get_map_base(self, _data):
        '''
        plot geo-VisualMap
        :param _data: gps对应的城市分布(dataframe)
        :return: 热点图
        '''
        register_url(self.echarts_url)
        c_map = Map(init_opts=opts.InitOpts(width="1000px", height="600px", page_title='', bg_color=''))
        c_map.add(series_name=self.series_name, data_pair=_data, maptype=self.country)
        c_map.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
        c_map.set_global_opts(title_opts=opts.TitleOpts(title="Map-VisualMap"),
                              visualmap_opts=opts.VisualMapOpts(max_= self.threshold_plot))
        return c_map

    def get_table_base(self,_data):
        '''
        plot geo-detail
        :param _data: gps对应的城市分布(dataframe)
        :return: 数据示例
        '''
        table = Table()
        headers = ["City name", "Number"]
        rows = _data
        table.add(headers, rows).set_global_opts(
            title_opts=opts.ComponentTitleOpts(title="Table-Details")
        )
        return table

    def get_city_list(self):
        df = pd.read_csv(os.path.join(CURRENT_PATH,'CITY.cg'))
        return df

    def run_geo_map(self,df_raw,web_open=False):
        df_gps = self.get_geo_reverse(df_raw)
        geo_data = df_gps.groupby("admin1")['lon'].count()
        self.plot_geo_summary(geo_data)
        if web_open:
            webbrowser.open(url='file://' + self.html_name, new=0, autoraise=True)

if __name__ == '__main__':
    df_raw = pd.DataFrame( data = [[-0.4302744, 35.5331974],
                                   [-1.21167  , 36.8973031],
                                   [ 0.5217472, 35.255192 ],
                                   [-2.2326433, 38.0769783],
                                   [-1.0625246, 34.4734754]],
                           columns=['latitude', 'longitude']
                           )
    geo_map = GeoMap()
    geo_map.run_geo_map(df_raw,web_open=True)


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-30 14:42
Desc:
    xgboost 模型评分卡
    xgboost 参数 https://blog.csdn.net/zc02051126/article/details/46711047
    score = self.score_offset + self.score_factor * np.log(odds)
'''

from pynovice.score_card.src.score_card import ScoreCardModel
import xgboost as xgb
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split

class XGBModel(ScoreCardModel):
    '''
        用于预测某一条预处理过的特征向量得分的方法
    :return:
        score_label: 最终预测的评分
        score_feature: 各个特征的得分字典
    '''
    missing_value: float

    def __init__(self,missing_value=-999.0,**kwargs):
        super().__init__()
        self.missing_value = missing_value
        self.param = self.load_xgb_param(kwargs)
        self.feature_columns = None
        self.model_ = None

    def load_xgb_param(self,kwargs):
        param={}
        # 1.一般参数:  这些参数用来控制XGBoost的整体通用功能，
        general_param={
            'booster': 'gbtree',  # [默认gbtree] 有两中模型可以选择gbtree和gblinear。gbtree使用基于树的模型进行提升计算，gblinear使用线性模型进行提升计算。
            'silent': 1, # [默认0]	是否开启静默模式，0为不开启，1为开启，开启后不打印运行时信息。
            'nthread': 1,  # [默认取最大线程数] 这个参数用来控制最大并行的线程数
        }
        # 2.booster的参数,取决于使用哪种booster, 这些参数是要重点调整的。
        booster_tree_param = {
            'eta': 0.1,  # [default=0.3] 学习率参数，就是原理中说的缩减，保证每一颗树对于结果的影响不太大，从而保证模型的效果。更新叶子节点权重时，乘以该系数，避免步长过大。参数值越大，越可能无法收敛。把学习率 eta 设置的小一些，小学习率可以使得后面的学习更加仔细。 典型值为0.01-0.2。
            'gamma': 0,  # [default=0] 一听这种希腊字母就知道是个系数，在树的叶子节点上作进一步分区所需的最小损失减少。越大，算法越保守。取值在[0,∞] 。通俗点讲就是，这个节点还划不划分，先看看损失减不减少了再说。同样需要cv调优。
            'max_depth': 3,  # [default=6] 这个没啥好说的，每棵树的最大深度，也是用来避免过拟合的，max_depth越大，模型会学到更具体更局部的样本，典型值3-10，要用cv调优。
            'min_child_weight': 1,  # [default=1] 大家对他的解释是决定最小叶子节点样本权重和，不太好理解。看了一些解释的文章，这个值可以理解为H值，还记得H值吗，就是损失函数对y(t-1)的二阶导数和，那么如果损失函数是平方函数（回归问题），这个就是1，如果是对数损失函数（分类问题），导数是a(1-a)的形式，a代表sigmoid函数，这样的话当y预测值非常大的时候，这个式子的值接近于0，这当然是不好的，因此你要给他设定一个阈值，小于这个阈值就不分裂了。现在可以解释了，这个值代表所有样本二阶导数的和，和上边说的叶子得分不是一个事，如果是回归问题实际代表样本个数，如果是分类问题实际代表a(1-a)所有样本计算值的加和。明白这个参数是啥以后，来看他是干嘛的，这个参数用于避免过拟合，当它的值较大时，可以避免模型学习到局部的特殊样本。举个栗子来说，对正负样本不均衡时的 0-1 分类而言，假设 h 在 0.01 附近，min_child_weight 为 1 意味着叶子节点中最少需要包含 100 个样本，实际是通过控制样本数来控制过拟合的。你们应该看出来这个值越小越容易过拟合，需要通过cv进行调整优化。
            # max_leaf_nodes 树上最大节点的数量，和上面的那个参数一样，如果定义了这个参数就会忽略掉max_depth参数，我们调优还是以max_depth为主吧。
            'max_delta_step': 0,  # [default=0] 这参数限制每棵树权重改变的最大步长。如果这个参数的值为0，那就意味着没有约束。如果它被赋予了某个正值，那么它会让这个算法更加保守。通常，这个参数不需要设置。但是当各类别的样本十分不平衡时，它对逻辑回归是很有帮助的。也就是说这个参数不用管啊。
            'subsample': 1,  # [default=1] 样本采样用的，用于训练模型的子样本占整个样本集合的比例,减小这个参数的值，算法会更加保守，避免过拟合，但是如果这个值设置得过小，它可能会导致欠拟合。典型值：0.5-1。既然有个范围，给他个面子cv调优一把吧。
            'colsample_bytree': 1,  # [default=1] 列采样，在建立树时对特征采样的比例，前面介绍过了，和设置缩减率一样是为了干嘛来着，是为了防止过拟合的，一般设置为： 0.5-1 ，也要用cv拟合。
            'tree_method': 'auto',  # [default=’auto’] 还记得我说过树的生成方法，有三个可选的值， {‘auto’, ‘exact’, ‘approx’} ，分别对应 贪心算法(小数据集)/近似算法(大数据集) 。
            #'scale_pos_weight': # 是用来调节正负样本不均衡问题的，用助于样本不平衡时训练的收敛（应该是增大了少数样本的学习率）。- 如何你仅仅关注预测问题的排序或者AUC指标，那么你尽管可以调节此参数。如果你希望得到真正的预测概率则不能够通过此参数来平衡样本
        }
        booster_linear_param = {
            'lambda': 1,   # [default=1] 又是个系数，这个是控制L2正则的，就是目标函数里的那个叶子节点得分前边的系数，用不用看你自己了。
            'alpha': 0,  # [default=0] 在建立树时对特征采样的比例有L2就有L1，用不用全凭自己了。
            'lambda_bias': 0,  # [default=0] 在偏置上的L2正则。缺省值为0（在L1上没有偏置项的正则，因为L1时偏置不重要）
        }
        # 学习目标参数,跟目标函数有关,控制学习的场景，例如在回归问题中会使用不同的参数控制排序
        object_param = {
            'objective': 'binary:logistic',  # [ default=reg:linear ] 定义学习任务及相应的学习目标，可选的目标函数如下,包含的函数还挺多，默认是线形的。此外你还可以选择：binary:logistic 二分类的逻辑回归，返回预测的概率(不是类别).
            'eval_metric':['error','auc'],  # [默认值取决于objective参数的取值] 也就是说怎么计算目标函数值，根据你目标函数的形式来，对于回归问题，默认值是rmse，对于分类问题，默认值是error。
            'base_score': 0.5,  # [ default=0.5 ]
            'seed':666,  # [ default=0 ] 随机数的种子。缺省值为0
        }
        param.update(general_param)
        param.update(booster_tree_param)
        param.update(object_param)
        param.update(kwargs)
        return param

    def train(self,train_feature, train_label, test_feature=pd.DataFrame(), test_label=pd.DataFrame(),eval=False,
              num_boost_round=200,early_stopping_rounds=50):
        '''

        :param train_feature: pd.DataFrame
        :param train_label:
        :param test_feature:
        :param test_label:
        :return:
        '''
        self.feature_columns = train_feature.columns
        if (test_feature.empty or test_label.empty):
            train_feature, test_feature, train_label, test_label = \
                train_test_split(train_feature, train_label, test_size=0.2, random_state=0)
        dtrain = xgb.DMatrix(data=train_feature, label=train_label, missing=self.missing_value)
        dtest = xgb.DMatrix(data=test_feature, label=test_label, missing=self.missing_value)
        evallist = [(dtrain, 'train'), (dtest, 'test')]  # 如果说eval_metric有很多个指标(evals)，那就以最后一个指标为准
        # 当logloss在设置early_stopping_rounds轮迭代之内，都没有提升的话，就stop。
        self.model_ = xgb.train(self.param, dtrain, num_boost_round=num_boost_round, evals=evallist,
                                early_stopping_rounds=early_stopping_rounds)
        if eval:
            # 模型评估
            print("training eval:")
            df_pre, _ = self.predict(train_feature)
            auc, _ = self.get_auc(df_pre,train_label)
            ks, _ = self.get_ks(df_pre, train_label)
            print("auc={}, ks={}".format(auc,ks['gap'].values[0]))

            print('testing eval:')
            df_pre, _ = self.predict(test_feature)
            self.plot_roc(df_pre, test_label, pre_target=1, save_path='.')
            self.plot_ks(df_pre, test_label, pre_target=1, save_path='.')
            self.plot_lift(df_pre, test_label, pre_target=1, save_path='.')

    def predict(self,x):
        '''
        :param x: list/datafame
        :return: array([proba]),array([score_label])
        '''
        dmatrix = xgb.DMatrix(np.array(x), feature_names=self.feature_columns, missing=self.missing_value)
        predict = self.model_.predict(dmatrix)
        score_label = np.array([self.probability_to_score(bad_prob=proba_i) for proba_i in predict])
        return predict, score_label


if __name__ == '__main__':
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 6, 8, 3, 2, 1, 5, 7],[10, 2, 3, 42, 534, 5, 53, 6, 83, 3, 42, 21, 25, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field1', 'field2', 'label']
    df_fields, df_label = df[['field1','field2']], df['label']
    lf = XGBModel(aa=34,missing_value=99)
    lf.train(df_fields,df_label,eval=False)
    print(lf.predict([[2,3]]))
    print(lf.predict(df_fields.head()))



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-28 17:59
Desc:
https://github.com/boredbird/woe

    iv最大值分箱 【迭代依次查找（合并分箱后iv最大的两个箱)进行合并】
    连续特征值观测值过多的可以先进行等频率分箱
    1.初始化阶段：首先按照属性值的大小进行排序（对于非连续特征，需要先做数值转换，比如转为坏人率，然后排序），然后每个属性值单独作为一组。

    2、合并阶段：
    （1）对每一对相邻的组，计算iv值。
    （2）根据计算的iv值，对其中最大的一对邻组合并为一组。
    （3）不断重复（1），（2）直到分组数达到一定的条件（如最小分组数5，最大分组数8）并且每箱都包含正负多个label。
对同一个变量划分bin越多，iv越高

未完成
'''

import pandas as pd
import numpy as np
from pynovice.score_card.src.data_boxing import frequence_blocks

def calc_woe(x,y,bad_target=1, eps=1e-4):
    '''
    :param x:
    :param y:
    :param bad_target:
    :param eps:
    :return:
        woe字典和此特征的iv值
    '''
    bin = np.unique(x)
    prob_bad = np.array([np.logical_and(x==val,y==bad_target).sum() for val in bin])/float(np.sum(y==bad_target))
    prob_good = np.array([np.logical_and(x==val,y!=bad_target).sum() for val in bin])/float(np.sum(y!=bad_target))
    woe = np.log(np.maximum(prob_good,eps)/np.maximum(prob_bad,eps))
    iv = (prob_good-prob_bad)*woe
    woe_dict = dict(zip(bin,woe))
    return woe_dict,iv

def iv_split(df_field, df_label,bad_target=1):
    df_field = df_field.sort_values()
    df_label = df_label[df_field.index]


    init_iv = calc_woe(df_field,df_label,bad_target)


    init_split = 0.5 * (df_field[1:] + df_field[:-1])



    def find_best_split(_df,init_split):
        # 找出最大iv的分割点
        max_iv = -1
        for i in init_split:
            df_left = _df[_df < i]
            df_right = _df[_df >= i]
            _df[df_left.index] = 0
            _df[df_right.index] = 1
            _,tmp_iv = calc_woe(_df, df_label, bad_target)
            if tmp_iv>max_iv:
                max_iv = tmp_iv
                best_idx = i
        return best_idx



def iv_blocks(df_field, df_label, maxCases=5, max_iterations=100):
    '''
    best_iv 分箱
    :param df_field: 需分箱的字段 Series
    :param df_label: 样本的标签 Series
    :param maxCases: 最大分箱数
    :param sample_threshold:  bins中最少样本数
    :return:
        分箱的eges
    '''
    if maxCases<2:
        return [-np.inf,np.inf]
    iterations_bins = maxCases + max_iterations
    if len(set(df_field)) > iterations_bins:  # 连续特征太多，进行初步分bins
        init_bins = frequence_blocks(x=df_field.to_list(), bins=iterations_bins)
        _bins = init_bins[1:-1]+[init_bins[-2]+10]
        df_field = pd.cut(df_field,bins=init_bins,labels=_bins)

    sample_threshold = df_field.shape[0] / 20  # 总体的5%(bins要大于5%)
    pass
    print('......')

    woe_dict,iv = calc_woe(df_field, df_label, bad_target=1)


if __name__ == '__main__':
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 3, 3, 3, 2, 1, 5, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field', 'label']
    df_field, df_label = df['field'], df['label']


    import lightgbm as lgb
    params = {
        'task': 'train',  # 执行的任务类型
        'boosting_type': 'gbrt',  # 基学习器
        'objective': 'lambdarank',  # 排序任务(目标函数)
        'metric': 'ndcg',  # 度量的指标(评估函数)
        'max_position': 10,  # @NDCG 位置优化
        'metric_freq': 1,  # 每隔多少次输出一次度量结果
        'train_metric': True,  # 训练时就输出度量结果
        'ndcg_at': [10],
        'max_bin': 255,  # 一个整数，表示最大的桶的数量。默认值为 255。lightgbm 会根据它来自动压缩内存。如max_bin=255 时，则lightgbm 将使用uint8 来表示特征的每一个值。
        'num_iterations': 200,  # 迭代次数，即生成的树的棵数
        'learning_rate': 0.01,  # 学习率
        'num_leaves': 31,  # 叶子数
        # 'max_depth':6,
        'tree_learner': 'serial',  # 用于并行学习，‘serial’： 单台机器的tree learner
        'min_data_in_leaf': 30,  # 一个叶子节点上包含的最少样本数量
        'verbose': 2  # 显示训练时的信息
    }
    train_data = lgb.Dataset(df_field,df_label)
    gbm = lgb.train(params, train_data, valid_sets=[train_data])

